#include <sstream>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <algorithm>
#include <cmath>
#include <map>
#include <limits>

#include "JobScheduling.h"

using namespace std;

/// index == neuron
/// value == userID
map<int,int> userIdLookupTable;

/// If someone hasn't used a program before, his "time since last run" is set to this.
const double MAX_TIME_SINCE_LAST_RUN = 365 * 24 * 60 * 60;
//const double MAX_WAIT_TIME = 30 * 24 * 60 * 60;

/// If a user previously hasn't used the application, his "previous run time" will be set to this.
double longestRunTime;

/// Parses a single line from a .swf file. If the line contains data, it is stored in &data and true is returned.
/// If it's a comment or invalid line false is returned.
bool parseJobLine(string line, JobData &jobData)
{
    stringstream ss(line);
    if(!(ss >> jobData.jobNumber)) return false; // 1.
    if(!(ss >> jobData.submitTime)) return false; // 2.
    if(!(ss >> jobData.waitTime)) return false; // 3.
    if(!(ss >> jobData.runTime)) return false; // 4.
    if(!(ss >> jobData.numberOfAllocatedProcessors)) return false; // 5.
    if(!(ss >> jobData.averageCPUTimeUsed)) return false; // 6.
    if(!(ss >> jobData.usedMemory)) return false; // 7.
    if(!(ss >> jobData.requestedNumberOfProcessors)) return false; // 8.
    if(!(ss >> jobData.requestedTime)) return false; // 9.
    if(!(ss >> jobData.requestedMemory)) return false; // 10.
    if(!(ss >> jobData.status)) return false; // 11.
    if(!(ss >> jobData.userID)) return false; // 12.
    if(!(ss >> jobData.groupID)) return false; // 13.
    if(!(ss >> jobData.applicationNumber)) return false; // 14.
    if(!(ss >> jobData.queueNumber)) return false; // 15.
    if(!(ss >> jobData.partitionNumber)) return false; // 16.
    if(!(ss >> jobData.precedingJobNumber)) return false; // 17.
    if(!(ss >> jobData.thinkTimeFromPrecedingJob)) return false; // 18

    return true;
}


vector<JobData> parseFile(string fileName)
{
    vector<JobData> jobs;
    ifstream file(fileName.c_str());

    if(!file.is_open())
        throw Exception("Error opening file");

    string line;
    while( getline(file, line) )
    {
        JobData tempJobData;
        if(parseJobLine(line, tempJobData))
            jobs.push_back(tempJobData);
    }

    file.close();

    return jobs;
}

vector<JobData> filterOutPartialJobs(const vector<JobData> &jobs)
{
    vector<JobData> filteredJobs;

    for(size_t i = 0; i < jobs.size(); ++i)
    {
        if(jobs[i].status == 0 || jobs[i].status == 1 || jobs[i].status == 5)
            filteredJobs.push_back(jobs[i]);
    }

    return filteredJobs;
}

vector<double> submitTime(const vector<JobData> &allJobs, int jobIndex)
{
    return vector<double>(1,allJobs[jobIndex].submitTime);
}

std::vector<double> waitTime(const std::vector<JobData> &allJobs, int jobIndex)
{
	return vector<double>(1,allJobs[jobIndex].waitTime);
}

vector<double> status(const vector<JobData> &allJobs, int jobIndex)
{
    return vector<double>(1, jobStatusNeuronValue(allJobs[jobIndex].status) );
}

vector<double> requestedNumberOfProcessors(const vector<JobData> &allJobs, int jobIndex)
{
    return vector<double>(1,allJobs[jobIndex].requestedNumberOfProcessors);
}

vector<double> requestedTime(const vector<JobData> &allJobs, int jobIndex)
{
    return vector<double>(1,allJobs[jobIndex].requestedTime);
}

vector<double> requestedMemory(const vector<JobData> &allJobs, int jobIndex)
{
    return vector<double>(1,allJobs[jobIndex].requestedMemory);
}

vector<double> userID(const vector<JobData> &allJobs, int jobIndex)
{
    int _userid = allJobs[jobIndex].userID;
    map<int,int>::const_iterator it = userIdLookupTable.find(_userid);
    int index=-1;
    if(it != userIdLookupTable.end())
    {
        index = it->second;
    }
    return createAttributeValues(userIdLookupTable.size(), index);
}

/// Returns the status of the last job of the same user.
/// If this is the first job, returns successful.
vector<double> lastStatusOfUser(const vector<JobData> &allJobs, int jobIndex)
{
	int thisUser = allJobs[jobIndex].userID;
    for(int i=jobIndex-1;i>=0;i--)
    {
        if(allJobs[i].userID == thisUser)
        	return vector<double>(1, jobStatusNeuronValue(allJobs[i].status) );
    }
    return vector<double>(1,JobData::SUCCESSFUL);
}

vector<double> lastStatusOfApplication(const vector<JobData> &allJobs, int jobIndex)
{
    int thisApplication = allJobs[jobIndex].applicationNumber;
    for(int i=jobIndex-1;i>=0;i--)
    {
        if(allJobs[i].applicationNumber == thisApplication)
            return vector<double>(1, jobStatusNeuronValue(allJobs[i].status) );
    }
    return vector<double>(1,JobData::SUCCESSFUL);
}

vector<double> lastStatusOfUserWithApplication(const vector<JobData> &allJobs, int jobIndex)
{
    int thisUser = allJobs[jobIndex].userID;
    int thisApplication = allJobs[jobIndex].applicationNumber;
    for(int i=jobIndex-1;i>=0;i--)
    {
        if(allJobs[i].userID == thisUser && allJobs[i].applicationNumber == thisApplication)
            return vector<double>(1, jobStatusNeuronValue(allJobs[i].status) );
    }
    return vector<double>(1,JobData::SUCCESSFUL);
}

// Returns 1 if last the users last job with the same application was cancelled, otherwise returns 0.
vector<double> lastJobOfUserWithApplicationWasCanceled(const vector<JobData> &allJobs, int jobIndex)
{
    int thisUser = allJobs[jobIndex].userID;
    int thisApplication = allJobs[jobIndex].applicationNumber;
    for(int i=jobIndex-1;i>=0;i--)
    {
        if(allJobs[i].userID == thisUser && allJobs[i].applicationNumber == thisApplication)
            return vector<double>(1, (double)allJobs[i].status == 5);
    }
    return vector<double>(1,JobData::SUCCESSFUL);
}


// Returns the run time of the previous job with the same user with the same application in seconds.
// If there was no job or it didn't have a specified run time, returns 0.
vector<double> lastRunTimeOfUserWithApplication(const vector<JobData> &allJobs, int jobIndex)
{
	double runTime = 0;
	int thisGuy = allJobs[jobIndex].userID;
	int thisApp = allJobs[jobIndex].applicationNumber;

	for(int i = jobIndex-1; i >= 0; --i)
	{
		if(allJobs[i].userID == thisGuy && allJobs[i].applicationNumber == thisApp)
		{
			runTime = allJobs[i].runTime;
			break;
		}
	}

	if(runTime == -1) runTime = 0;

	return vector<double>(1, runTime);

}

// Returns the average CPU time of the previous job with the same user with the same application in seconds.
// If there was no job or it didn't have a specified average CPU time, returns 0.
vector<double> lastAverageCPUTimeOfUserWithApplication(const vector<JobData> &allJobs, int jobIndex)
{
	double avgCPUTime = 0;
	int thisGuy = allJobs[jobIndex].userID;
	int thisApp = allJobs[jobIndex].applicationNumber;

	for(int i = jobIndex-1; i >= 0; --i)
	{
		if(allJobs[i].userID == thisGuy && allJobs[i].applicationNumber == thisApp)
		{
			avgCPUTime = allJobs[i].averageCPUTimeUsed;
			break;
		}
	}

	if(avgCPUTime == -1) avgCPUTime = 0;

	return vector<double>(1, avgCPUTime);

}

// Returns the how much time has passed since user last used that application. Time is in seconds.
// If the user hasn't used it yet
vector<double> elapsedTimeSinceUserLastUsedApplication(const vector<JobData> &allJobs, int jobIndex)
{
	double elapsed = numeric_limits<double>::max();
	double now = allJobs[jobIndex].submitTime;
	int thisGuy = allJobs[jobIndex].userID;
	int thisApp = allJobs[jobIndex].applicationNumber;

	for(int i = jobIndex-1; i >= 0; --i)
	{
		if(allJobs[i].userID == thisGuy && allJobs[i].applicationNumber == thisApp)
		{
			elapsed = now - allJobs[i].submitTime;
			break;
		}
	}

	elapsed = min(elapsed, MAX_TIME_SINCE_LAST_RUN);

	return vector<double>(1, elapsed);
}

vector<double> elapsedTimeSinceUserLastUsedCluster(const vector<JobData> &allJobs, int jobIndex)
{
	double elapsed = numeric_limits<double>::max();
	double now = allJobs[jobIndex].submitTime;
	int thisGuy = allJobs[jobIndex].userID;

	for(int i = jobIndex-1; i >= 0; --i)
	{
		if(allJobs[i].userID == thisGuy)
		{
			elapsed = now - allJobs[i].submitTime;
			break;
		}
	}

	elapsed = min(elapsed, MAX_TIME_SINCE_LAST_RUN);

	return vector<double>(1, elapsed);

}



/// Converts a status value from a log file value (1 - successful, 0,5 - failed), to a neuron value (1 - failed, 0 - successful).
double jobStatusNeuronValue(int status)
{
    if(status == 1) return JobData::SUCCESSFUL;
    else return JobData::FAILED;
}


void createTrainingFile(const vector<JobData> &allJobs, const vector<AttributeFunction> &inputAttributes, const vector<AttributeFunction> &outputAttributes, string outputFileName, double trainingSetSize, double validationSetSize, double testSetSize)
{
    ofstream file(outputFileName.c_str());
    if(!file.is_open())
        throw Exception("Error opening file");

    // training set, validation set and test set ratios
    file << trainingSetSize << " " << validationSetSize << " " << testSetSize << endl;

    // input and output size
    int inputSz = inputSize(allJobs, inputAttributes);
    int outputSz = outputSize(allJobs, outputAttributes);
    file << inputSz << " " << outputSz << endl;

    // Preprocessing

    // find longest run time
    for(size_t i = 0; i < allJobs.size(); ++i)
    	longestRunTime = std::max(allJobs[i].runTime, longestRunTime);

    // create lookup tables
    createUserIdLookupTable(allJobs);




    // ranges of the samples
    vector<pair<double, double> > ranges(inputSz + outputSz, pair<double,double>(numeric_limits<double>::max(), numeric_limits<double>::min()));

    int positives = 0, negatives = 0;

    // create samples
    vector<vector<double> > samples;
    for(size_t i=0;i<allJobs.size();i++)
    {
        vector<double> sample = createTrainingSample(allJobs, i, inputAttributes, outputAttributes);
        samples.push_back(sample);

        // check ranges
        for(size_t j = 0; j < sample.size(); ++j)
		{
        	ranges[j].first = min(ranges[j].first, sample[j]);
        	ranges[j].second = max(ranges[j].second, sample[j]);
		}

        // positives, negatives
        double out = sample.back();
        if(out > 0.5) positives++;
        else negatives++;
    }

    cout << "positives = " << positives << "\tnegatives= " << negatives << endl;

    // write ranges to file
    double increaseRange = 0.25; // range is increased by half of that on each side
    for(size_t i = 0; i < ranges.size(); ++i)
	{
    	double range = ranges[i].second - ranges[i].first;
    	if(abs(range) < 1e-200) range = 1; // if all values are the same // vjerojatno bi trebo nes ljepse od 1e-200
    	double delta = (increaseRange * range) / 2;
    	ranges[i].first -= delta;
    	ranges[i].second += delta;

    	file << ranges[i].first << " " << ranges[i].second << endl;
	}

    // write samples to file
    for(size_t i = 0; i < samples.size(); ++i)
	{
        for(size_t j=0; j<samples[i].size();j++)
            file << samples[i][j] << "\t";
        file << '\n';
	}

    file.close();
}

/// Changes the ratio of positives to all samples in the training and validation set by deleting the negatives.
/// If the there already are enough positives, nothing is changed.
/// Test set is unchanged.
int changeTrainingSamplePositivesRatio(std::vector<JobData> &allJobs, int trainingAndValidationSize, double positivesRatio)
{
	int positives = 0;
	for(int i = 0; i < trainingAndValidationSize; ++i)
	{
		if( jobStatusNeuronValue(allJobs[i].status) == JobData::FAILED)
			positives++;
	}

	int trainingAndValidFinal = (positivesRatio > 0)? floor(positives / positivesRatio + 0.5) : trainingAndValidationSize;
	int negativesToDelete = trainingAndValidationSize-trainingAndValidFinal;

	cout << "now = " << trainingAndValidationSize << "  final =" <<  trainingAndValidFinal << endl;

	while(trainingAndValidationSize > trainingAndValidFinal)
	{
		int indexToRemove = rand()%trainingAndValidationSize;
		if(jobStatusNeuronValue(allJobs[indexToRemove].status) == JobData::SUCCESSFUL)
		{
			allJobs.erase(allJobs.begin() + indexToRemove);
			trainingAndValidationSize--;
		}
	}

	cout << "Negatives removed = " << negativesToDelete << endl;

	return trainingAndValidationSize;
}

void createUserIdLookupTable(const vector<JobData> &allJobs)
{
    int currentNeuronIndex = 0;
    for(size_t i=0;i<allJobs.size();i++)
    {
        int val = allJobs[i].userID;
        if(val >= 0)
        {
            pair<map<int,int>::iterator,bool> ret;
            ret = userIdLookupTable.insert(pair<int,int>(val, currentNeuronIndex)); // if element exists, it is not inserted
            if(ret.second == true) // it is inserted
                currentNeuronIndex++;
        }
    }
    cout << "LUT size=" << userIdLookupTable.size() << endl;


}

/// Creates a vector of doubles which represent a single line in a training file - a sample.
vector<double> createTrainingSample(const vector<JobData> &allJobs, int jobIndex, const vector<AttributeFunction> &inputAttributes, const vector<AttributeFunction> &outputAttributes)
{
    vector<double> sample;
    for(size_t i=0;i<inputAttributes.size();i++)
    {
        vector<double> attribute = inputAttributes[i](allJobs, jobIndex);
        sample.insert(sample.end(), attribute.begin(), attribute.end());
    }

    for(size_t i=0;i<outputAttributes.size();i++)
    {
        vector<double> attribute = outputAttributes[i](allJobs, jobIndex);
        sample.insert(sample.end(), attribute.begin(), attribute.end());
    }
    return sample;
}

/// Returns the number of input values(neurons).
int inputSize(const vector<JobData> &allJobs, const vector<AttributeFunction> &inputAttributes)
{
    return (createTrainingSample(allJobs, 0, inputAttributes,vector<AttributeFunction>())).size();
}

/// Returns the number of output values(neurons).
int outputSize(const vector<JobData> &allJobs, const vector<AttributeFunction> &outputAttributes)
{
    return (createTrainingSample(allJobs, 0, vector<AttributeFunction>(), outputAttributes)).size();
}

/// Creates a vector size of numberOfNeurons with all neurons set to 0 except activeNeuron which is set to 1.
/// If active neuron is -1, no one is active.
vector<double> createAttributeValues(int numberOfNeurons, int activeNeuron)
{
    vector<double> values(numberOfNeurons,0);
    if(activeNeuron < -1 || activeNeuron >= numberOfNeurons)
        throw Exception("Wrong active neuron");
    if(activeNeuron != -1)
        values[activeNeuron] = 1;
    return values;
}
