
#include <iostream>
#include <string>
#include <ctime>
#include <cstdlib>
#include <cstdio>

#include <CL/opencl.h>

#include "NN.h"
#include "util.h"
#include "NN_factory.h"
#include "NN_statistics.h"

using namespace std;

void sample_size_statistics(int start, int increment, int end, int numInterations, string fileName);
void num_neurons_in_layer_statistics(int start, int increment, int end, int numInterations, string fileName, int numSamples);
void num_layers_statistics(int start, int increment, int end, int neuronsInLayer, int numInterations, string fileName, int numSamples);
void BP_num_neurons_in_layer_statistics(int start, int increment, int end, int numHiddenLayers, int numInterations, string fileName, int numSamples);

void print_usage_info()
{
    cout << "Usage: ./NeuralNetwork <NN_type> <nnet_config_file>  <training_data_file>  <algorithm_parameters_file> [<network_file>]" << endl;
    cout << "       ./NeuralNetwork <network_file>" << endl;
}

int main(int argc, char *argv[])
{
    //srand(0);
    srand(time(NULL));

    //openclPrintAllDevicesInfo();

    // parse command line arguments
    NN *nn = NULL;

    if(argc == 6 || argc == 5) // train nn
    {
        string NN_type = argv[1];
        string nnet_config_file = argv[2];
        string training_data_file = argv[3];
        string algorithm_params_file  = argv[4];

        // create neural network
        if(NN_type == "PSO") nn = new NN_PSO(nnet_config_file, training_data_file, algorithm_params_file);
        else if(NN_type == "PSO_local_GPU") nn = new NN_PSO_OpenCL_GPU_local(nnet_config_file, training_data_file, algorithm_params_file);
        else if(NN_type == "PSO_local_CPU") nn = new NN_PSO_OpenCL_CPU_local(nnet_config_file, training_data_file, algorithm_params_file);
        else if(NN_type == "PSO_private_GPU") nn = new NN_PSO_OpenCL_GPU_private(nnet_config_file, training_data_file, algorithm_params_file);
        else if(NN_type == "PSO_private_CPU") nn = new NN_PSO_OpenCL_CPU_private(nnet_config_file, training_data_file, algorithm_params_file);
        else if(NN_type == "BP") nn = new NN_BP(nnet_config_file, training_data_file, algorithm_params_file);
        else if(NN_type == "BP_GPU") nn = new NN_BP_OpenCL_GPU(nnet_config_file, training_data_file, algorithm_params_file);
        else if(NN_type == "BP_CPU") nn = new NN_BP_OpenCL_CPU(nnet_config_file, training_data_file, algorithm_params_file);
        else error("Wrong NN type");

        nn->printInfo();

        cout << "Training..." << endl;
        nn->train();
        cout << "Network trained" << endl;

        cout << "Test set MSE= " << nn->testMSE() << endl;
        cout << "Test set TPR= " << nn->testTPR() << endl;

        // save to file
        if(argc == 6)
        {
            string NN_weights_file = argv[5];
            nn->saveWeightsToFile(NN_weights_file);
        }
    }
    else if(argc == 2) // load nn from file
    {
        string NN_weights_file = argv[1];
        nn = new NN_reloaded(NN_weights_file);
        nn->printInfo();
    }
    else
    {
        print_usage_info();
        error("Wrong number of arguments");
    }


    // use network
    while(1)
    {
        cout << "Input:";
        vector<float> in(nn->getInputSize());
        for(int i=0;i<nn->getInputSize();i++)
            cin >> in[i];
        cout << "Output: ";
        vector<float> out = (*nn)(in);
        for(size_t i=0;i<out.size();i++)
            cout << out[i] << " ";
        cout << endl;
    }

    delete nn;

    // statistics
    // PSO
    //sample_size_statistics(500,500,5000, 5,"sample_size_timing_xxxs");

    //num_neurons_in_layer_statistics(3,3,60,20,"neurons_in_layer_60", 10000);

    //num_layers_statistics(1, 1, 30, 5, 3, "num_layers_30_5", 10000);

    // BP
    //BP_num_neurons_in_layer_statistics(5200, 200, 7000, 3, 1, "BP_num_layers_8000_3", 10);

    // Job Scheduling
    //NN_statistics::Job_Scheduling_statistics("_allAtrributes_0.5.txt", "nnet_config_just_all_attributes.txt", "PSO_params_all_attributes.txt");

    return 0;
}


void sample_size_statistics(int start, int increment, int end, int numInterations, string fileName)
{
    NN_statistics::NN_sample_size_timings(NN_statistics::PSO_CPU, start, increment, end, fileName + "_PSO CPU.csv", numInterations, "nn_config_sample_size_timings.txt", "samples_CTC-SP2-1996-2.1.txt", "PSO_params_samples_timings.txt");
    NN_statistics::NN_sample_size_timings(NN_statistics::PSO_OpenCL_local_GPU, start, increment, end, fileName + "_PSO local GPU.csv", numInterations, "nn_config_sample_size_timings.txt", "samples_CTC-SP2-1996-2.1.txt", "PSO_params_samples_timings.txt");
    NN_statistics::NN_sample_size_timings(NN_statistics::PSO_OpenCL_private_GPU, start, increment, end, fileName + "_PSO private GPU.csv", numInterations, "nn_config_sample_size_timings.txt", "samples_CTC-SP2-1996-2.1.txt", "PSO_params_samples_timings.txt");
    NN_statistics::NN_sample_size_timings(NN_statistics::PSO_OpenCL_local_CPU, start, increment, end, fileName + "_PSO local CPU.csv", numInterations, "nn_config_sample_size_timings.txt", "samples_CTC-SP2-1996-2.1.txt", "PSO_params_samples_timings.txt");
    NN_statistics::NN_sample_size_timings(NN_statistics::PSO_OpenCL_private_CPU, start, increment, end, fileName + "_PSO private CPU.csv", numInterations, "nn_config_sample_size_timings.txt", "samples_CTC-SP2-1996-2.1.txt", "PSO_params_samples_timings.txt");
}

void num_neurons_in_layer_statistics(int start, int increment, int end, int numInterations, string fileName, int numSamples)
{
    int numInputs = 9;
    int numOutputs = 1;
    NN_statistics::NN_num_neurons_in_layer_timings(NN_statistics::PSO_CPU, start, increment, end, fileName + "_PSO CPU.csv", numInterations, numInputs, numOutputs, "neurons_in_layer_CTC-SP2-1996-2.1.txt", "PSO_params_neurons_in_layer_timings.txt", numSamples);
    NN_statistics::NN_num_neurons_in_layer_timings(NN_statistics::PSO_OpenCL_local_GPU, start, increment, end, fileName + "_PSO local GPU.csv", numInterations, numInputs, numOutputs, "neurons_in_layer_CTC-SP2-1996-2.1.txt", "PSO_params_neurons_in_layer_timings.txt", numSamples);
    NN_statistics::NN_num_neurons_in_layer_timings(NN_statistics::PSO_OpenCL_private_GPU, start, increment, end, fileName + "_PSO private GPU.csv", numInterations, numInputs, numOutputs, "neurons_in_layer_CTC-SP2-1996-2.1.txt", "PSO_params_neurons_in_layer_timings.txt", numSamples);
    NN_statistics::NN_num_neurons_in_layer_timings(NN_statistics::PSO_OpenCL_local_CPU, start, increment, end, fileName + "_PSO local CPU.csv", numInterations, numInputs, numOutputs, "neurons_in_layer_CTC-SP2-1996-2.1.txt", "PSO_params_neurons_in_layer_timings.txt", numSamples);
    NN_statistics::NN_num_neurons_in_layer_timings(NN_statistics::PSO_OpenCL_private_CPU, start, increment, end, fileName + "_PSO private CPU.csv", numInterations, numInputs, numOutputs, "neurons_in_layer_CTC-SP2-1996-2.1.txt", "PSO_params_neurons_in_layer_timings.txt", numSamples);
}

void num_layers_statistics(int start, int increment, int end, int neuronsInLayer, int numInterations, string fileName, int numSamples)
{
    int numInputs = 9;
    int numOutputs = 1;
    NN_statistics::NN_num_layers_timings(NN_statistics::PSO_CPU, start, increment, end, neuronsInLayer, fileName + "_PSO CPU.csv", numInterations, numInputs, numOutputs, "num_layers_CTC-SP2-1996-2.1.txt", "PSO_params_num_layers_timings.txt", numSamples);
    NN_statistics::NN_num_layers_timings(NN_statistics::PSO_OpenCL_local_GPU, start, increment, end, neuronsInLayer, fileName + "_PSO local GPU.csv", numInterations, numInputs, numOutputs, "num_layers_CTC-SP2-1996-2.1.txt", "PSO_params_num_layers_timings.txt", numSamples);
    NN_statistics::NN_num_layers_timings(NN_statistics::PSO_OpenCL_private_GPU, start, increment, end, neuronsInLayer, fileName + "_PSO private GPU.csv", numInterations, numInputs, numOutputs, "num_layers_CTC-SP2-1996-2.1.txt", "PSO_params_num_layers_timings.txt", numSamples);
    NN_statistics::NN_num_layers_timings(NN_statistics::PSO_OpenCL_local_CPU, start, increment, end, neuronsInLayer, fileName + "_PSO local CPU.csv", numInterations, numInputs, numOutputs, "num_layers_CTC-SP2-1996-2.1.txt", "PSO_params_num_layers_timings.txt", numSamples);
    NN_statistics::NN_num_layers_timings(NN_statistics::PSO_OpenCL_private_CPU, start, increment, end, neuronsInLayer, fileName + "_PSO private CPU.csv", numInterations, numInputs, numOutputs, "num_layers_CTC-SP2-1996-2.1.txt", "PSO_params_num_layers_timings.txt", numSamples);
}

void BP_num_neurons_in_layer_statistics(int start, int increment, int end, int numHiddenLayers, int numInterations, string fileName, int numSamples)
{
    int numInputs = 9;
    int numOutputs = 1;
    NN_statistics::NN_num_neurons_in_layer_timings(NN_statistics::BP_CPU, start, increment, end, fileName + "_BP CPU.csv", numInterations, numInputs, numOutputs, "BP_num_neur_in_layers_10_CTC-SP2-1996-2.1.txt", "BP_params.txt", numSamples, numHiddenLayers);
    NN_statistics::NN_num_neurons_in_layer_timings(NN_statistics::BP_OpenCL_GPU, start, increment, end, fileName + "_BP OpenCL GPU.csv", numInterations, numInputs, numOutputs, "BP_num_neur_in_layers_10_CTC-SP2-1996-2.1.txt", "BP_params.txt", numSamples, numHiddenLayers);
    NN_statistics::NN_num_neurons_in_layer_timings(NN_statistics::BP_OpenCL_CPU, start, increment, end, fileName + "_BP OpenCL CPU.csv", numInterations, numInputs, numOutputs, "BP_num_neur_in_layers_10_CTC-SP2-1996-2.1.txt", "BP_params.txt", numSamples, numHiddenLayers);
}

